"""Bond Valuation functions

Important functions are `bond_price`, `bond_yield` and `bond_duration`
for computing the price, yield to maturity and duration of one
or more bonds.

The settlement date and maturity date can be given as dates and the
software calculates the time to maturity (and coupon dates) using a
`daycount` convention to calculate year_fractions.
For any `daycount` other than simple counting of days
(ACT/365 in ISDA terminology), this packages relies on the
`isda_daycounters` module that can be downloaded from
<https://github.com/miradulo/isda_daycounters>

Maturity can alternatively be given in years by setting `settle`
to `None`. The trade/settlement date is then year 0.

"""
from numpy import array, where, vectorize, float64, ceil
import pandas as pd
from bond_pricing.utils import (
    newton_wrapper, edate, dict_to_dataframe)
from bond_pricing.present_value import pvaf, equiv_rate


def _bond_coupon_periods_0(settle=None, mat=1, freq=2, daycount=None):
    r"""Unvectorized form of bond_coupon_periods.

    Please see its documentation

    """
    if settle is not None:
        settle = pd.to_datetime(settle)
        mature = pd.to_datetime(mat)
        # approximate number of full coupon periods left
        n = int(freq * (mature - settle).days / 360)
        # the divisor of 360 guarantees this is an overestimate
        # we keep reducing n till it is right
        while (edate(mature, -n * 12 / freq) <= settle):
            n -= 1
        next_coupon = edate(mature, -n * 12 / freq)
        n += 1  # n is now number of full coupons since previous coupon
        prev_coupon = edate(mature, -n * 12 / freq)
        # old wrong code: prev_coupon = edate(next_coupon, -12/freq)
        # if prev_coupon != settle:
        #     # we are in the middle of a coupon period
        #     # no of coupons is One PLUS No of full coupon periods
        #     n += 1
        if daycount is None:
            daycounter = default_daycounter
        else:
            assert daycount in daycounters, (
                "Unknown daycount {:}. {:}".format(
                    daycount,
                    "isda_daycounters not available"
                    if no_isda else ""))
            daycounter = daycounters[daycount]
        discounting_fraction = daycounter.year_fraction(
            settle, next_coupon) * freq
        accrual_fraction = daycounter.year_fraction(
            prev_coupon, settle) * freq
    else:
        n = ceil(mat*freq)
        accrual_fraction = n - freq * mat
        discounting_fraction = 1 - accrual_fraction
        next_coupon = None
        prev_coupon = None
    if accrual_fraction == 1:
        # We are on coupon date. Assume that bond is ex-interest
        # Remove today's coupon
        # This affects dirty price and accrued interest
        # but not clean price
        discounting_fraction += 1
        accrual_fraction -= 1
    return (n, discounting_fraction, accrual_fraction,
            next_coupon, prev_coupon)


_bond_coupon_periods = vectorize(_bond_coupon_periods_0)


def bond_coupon_periods(settle=None, mat=1, freq=2, daycount=None,
                        return_dataframe=False):
    r""" Compute no of coupon, coupon dates and fractions

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    mat : int or date
          Maturity date or if settle is None, maturity in years
    freq : int
          Coupon frequency
    daycount : daycounter
          This class has day_count and year_fraction methods
    return_dataframe : bool
         whether to return pandas DataFrame instead of dict

    Returns
    -------
    dict
        n : No of coupons left

        discounting_fraction: Fraction of coupon period to next coupon

        accrual_fraction: Fraction of coupon period to previous coupon

        next_coupon: Next coupon date (None if settle is None)

        prev_coupon Previous coupon date (None if settle is None)

    Examples
    --------
    >>> bond_coupon_periods(
    ... settle='2020-03-13', mat='2030-01-01', freq=2, daycount=None
    ... )  # doctest: +NORMALIZE_WHITESPACE
    {'n': 20,
    'discounting_fraction': 0.6,
    'accrual_fraction': 0.4,
    'next_coupon': Timestamp('2020-07-01 00:00:00'),
    'prev_coupon': Timestamp('2020-01-01 00:00:00')}

    >>> bond_coupon_periods(
    ... mat=10.125, freq=2, daycount=None
    ... )  # doctest: +NORMALIZE_WHITESPACE
    {'n': 21.0,
    'discounting_fraction': 0.25,
    'accrual_fraction': 0.75,
    'next_coupon': None,
    'prev_coupon': None}

    """
    res = array(_bond_coupon_periods(settle=settle, mat=mat,
                                     freq=freq, daycount=daycount))
    if len(res.shape) > 1:
        result = dict(n=res[0, :],
                      discounting_fraction=res[1, :],
                      accrual_fraction=res[2, :],
                      next_coupon=res[3, :],
                      prev_coupon=res[4, :],)
    else:
        result = dict(n=res[0][()],
                      discounting_fraction=res[1][()],
                      accrual_fraction=res[2][()],
                      next_coupon=res[3][()],
                      prev_coupon=res[4][()],)
    if return_dataframe:
        return dict_to_dataframe(result)
    else:
        return result


def bond_price_breakup(settle=None, cpn=0, mat=1, yld=0, freq=2,
                       comp_freq=None, face=100, redeem=None,
                       daycount=None, return_dataframe=False):
    r"""Compute clean/dirty price & accrued_interest of coupon bond using YTM

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    yld : float
          The yield to maturity in decimal
    freq : int
          Coupon frequency
    comp_freq : int
          Compounding frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods
    return_dataframe : bool
         whether to return pandas DataFrame instead of dict

    Returns
    -------
    dict
         dirty: dirty price of the bond

         accrued: accrued interest

         clean: clean price of the bond

         next_coupon: Next coupon date (only if settle is None)

         prev_coupon: Previous coupon date (only if settle is None)

    Examples
    --------
    >>> bond_price_breakup(
    ... settle="2012-04-15", mat="2022-01-01", cpn=8e-2, yld=8.8843e-2,
    ... freq=1)  # doctest: +NORMALIZE_WHITESPACE
    {'DirtyPrice': 96.64322827099208,
    'AccruedInterest': 2.311111111111111,
    'CleanPrice': 94.33211715988098,
    'NextCoupon': Timestamp('2013-01-01 00:00:00'),
    'PreviousCoupon': Timestamp('2012-01-01 00:00:00')}

    >>> bond_price_breakup(
    ... mat=10.25, cpn=8e-2, yld=9e-2,
    ... freq=2)  # doctest: +NORMALIZE_WHITESPACE
    {'DirtyPrice': 95.37373582338677,
    'AccruedInterest': 2.0,
    'CleanPrice': 93.37373582338677,
    'NextCoupon': None,
    'PreviousCoupon': None}

    >>> bond_price_breakup(
    ... settle="2012-04-15", mat="2022-01-01", cpn=8e-2, yld=8.8843e-2,
    ... freq=[1,2,4])  # doctest: +NORMALIZE_WHITESPACE
    {'DirtyPrice': array([96.64322827099208, 96.61560601490777,
     94.59488828646788], dtype=object),
    'AccruedInterest': array([2.311111111111111, 2.311111111111111,
    0.3111111111111111], dtype=object),
    'CleanPrice': array([94.33211715988098, 94.30449490379667,
    94.28377717535678], dtype=object),
    'NextCoupon': array([Timestamp('2013-01-01 00:00:00'),
                        Timestamp('2012-07-01 00:00:00'),
                        Timestamp('2012-07-01 00:00:00')],
    dtype=object),
    'PreviousCoupon': array([Timestamp('2012-01-01 00:00:00'),
                            Timestamp('2012-01-01 00:00:00'),
                            Timestamp('2012-04-01 00:00:00')],
                            dtype=object)}
    >>> bond_price_breakup(
    ... settle="2012-04-15", mat="2022-01-01", cpn=8e-2, yld=8.8843e-2,
    ... freq=[1,2,4],
    ... return_dataframe=True)
      DirtyPrice AccruedInterest CleanPrice NextCoupon PreviousCoupon
    0  96.643228        2.311111  94.332117 2013-01-01     2012-01-01
    1  96.615606        2.311111  94.304495 2012-07-01     2012-01-01
    2  94.594888        0.311111  94.283777 2012-07-01     2012-04-01

    """
    # None can make comp_freq an array of objects
    # Since log needs float, we use .astype(float64)
    freq, cpn = array(freq), array(cpn)
    comp_freq = where(comp_freq is None, freq, comp_freq).astype(float64)
    # find the equivalent yield that matches the coupon frequency
    yld = equiv_rate(yld, comp_freq, freq)
    redeem = where(redeem is None, face, redeem)
    red_by_face = redeem / face
    res = bond_coupon_periods(settle, mat, freq, daycount)
    # compounding factor from previous coupon date to settlement date
    fractional_CF = (1 + yld/freq)**res['accrual_fraction']
    # calculate PV as at previous coupon date and compound to today
    dirty = fractional_CF * (
        # PV of coupons as at previous coupon date
        cpn/freq * pvaf(yld/freq, res['n'])
        # PV of redemption as at previous coupon date
        + red_by_face * (1 + yld/freq)**-res['n'])
    accrued = cpn/freq * res['accrual_fraction']
    clean = dirty - accrued
    result = dict(DirtyPrice=(face*dirty)[()],
                  AccruedInterest=(face*accrued)[()],
                  CleanPrice=(face*clean)[()],
                  NextCoupon=res['next_coupon'],
                  PreviousCoupon=res['prev_coupon'])
    if return_dataframe:
        return dict_to_dataframe(result)
    else:
        return result


def bond_price(settle=None, cpn=0, mat=1,
               yld=0, freq=2, comp_freq=None,
               face=100, redeem=None, daycount=None):
    """ Compute clean price of coupon bond using YTM

    This is a wrapper for bond_price_breakup that extracts
    and returns only the clean price.

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    yld : float
          The yield to maturity in decimal
    freq : int
          Coupon frequency
    comp_freq : int
          Compounding frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods

    Returns
    -------
    float
        clean price of the bond

    Examples
    --------
    >>> bond_price(settle="2012-04-15", mat="2022-01-01", cpn=8e-2,
    ... yld=8.8843e-2, freq=1)
    94.33211715988098

    >>> bond_price(mat=10.25, cpn=8e-2, yld=9e-2, freq=2)
    93.37373582338677

    >>> bond_price(settle="2012-04-15", mat="2022-01-01", cpn=8e-2,
    ...            yld=8.8843e-2,freq=[1, 2, 4])
    array([94.33211715988098, 94.30449490379667, 94.28377717535678],
          dtype=object)

    >>> bond_price(settle = "2021-01-01", mat = "2031-01-01",
    ...            yld = 1e-2, freq = 2, cpn = 5e-2)
    137.9748382933389
    """
    return bond_price_breakup(
        settle=settle, cpn=cpn, mat=mat, yld=yld, freq=freq,
        comp_freq=comp_freq, face=face, redeem=redeem,
        daycount=daycount)['CleanPrice']


def bond_duration(settle=None, cpn=0, mat=1, yld=0, freq=2,
                  comp_freq=None, face=100, redeem=None,
                  daycount=None, modified=False):
    r"""

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    yld : float
          The yield to maturity in decimal
    freq : int
          Coupon frequency
    comp_freq : int
          Compounding frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods
    modified : bool
          Whether to return modified duration

    Returns
    -------
    float
         The duration (modified duration if modified is True)

    Examples
    --------
    >>> bond_duration(settle="2012-04-15", mat="2022-01-01", cpn=8e-2,
    ...               yld=8.8843e-2)
    6.678708669753968

    >>> bond_duration(settle="2012-04-15", mat="2022-01-01", cpn=8e-2,
    ...               yld=8.8843e-2, modified=True)
    6.394648779016871

    >>> bond_duration(settle="2012-04-15", mat="2022-01-01", cpn=8e-2,
    ...               yld=[7e-2, 8.8843e-2])
    array([6.88872548, 6.67870867])

    """
    # consider a bond which has just paid a coupon
    # decompose it as a portfolio of three positions
    # Position                 PV           Duration_minus_(1+y)/y
    # --------                 --           ----------------------
    # perpetuity of c       c/y = cF/Fy                0
    # short perpetuity     -c/y/F = -c/Fy              T
    # Redemption of R        R/F = Ry/Fy            T-(1+y)/y
    # where F is Future Value Factor (1+y)^T
    # Portfolio Duration is (1+y)/y - N/D where
    # N = cT - RyT + R(1+y) = R[1+y + T(c/R-y)]
    # D = cF - c + Ry = R[c(F-1)/R + y]
    # where Fy has been eliminated from both N and D
    # Eliminating R gives
    # N = 1+y + T(c/R-y)
    # D = c(F-1)/R + y
    # we compute duration in coupon periods as on previous coupon date
    freq, cpn = array(freq), array(cpn)
    comp_freq = where(comp_freq is None, freq, comp_freq).astype(float64)
    # find the equivalent yield that matches the coupon frequency
    yld = equiv_rate(yld, comp_freq, freq)
    redeem = where(redeem is None, face, redeem)
    R = redeem / face
    res = bond_coupon_periods(settle=settle, mat=mat, freq=freq,
                              daycount=daycount)
    y = yld/freq
    c = cpn/freq
    T = res['n']
    # T += where(res['accrual_fraction'] > 0, 1, 0)
    F = (1+y)**T
    dur = (1+y)/y - (1+y + T*(c/R-y)) / (c*(F-1)/R + y)
    # now we subtract the fractional coupon period
    dur -= res['accrual_fraction']
    # then we convert to years
    dur /= freq
    return where(modified, dur/(1+y), dur)[()]


def _bond_yield_0(settle=None, cpn=0, mat=1, price=100, freq=2, comp_freq=None,
                  face=100, redeem=None, daycount=None, guess=10e-2):
    r"""Unvectorized form of bond_yield.

    Please see its documentation

    """
    return newton_wrapper(
        lambda yld: bond_price(
            settle, cpn, mat, yld, freq, comp_freq, face, redeem,
            daycount) - price,
        guess)


def bond_yield(settle=None, cpn=0, mat=1, price=100, freq=2, comp_freq=None,
               face=100, redeem=None, daycount=None, guess=10e-2):
    r"""Find the yield to maturity of one or more bonds

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    price :
          The price of the bond
    freq : int
          Coupon frequency
    comp_freq : int
          Compounding frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods
    guess : float
          Initial guess of the yield for the root finder

    Returns
    -------
    float :
          The yield to maturity of the bond. np.nan if the root finder failed.

    Examples
    --------

    >>> bond_yield(settle="2012-04-15", mat="2022-01-01", cpn=8e-2,
    ... price=94.33, freq=1)
    0.08884647275135965

    >>> bond_yield(mat=10.25, cpn=8e-2, price=93.37, freq=2)
    0.09000591604105035

    >>> bond_yield(settle="2012-04-15", mat="2022-01-01", cpn=8e-2,
    ... price=[93, 94, 95], freq=1)
    array([0.09104904, 0.08938905, 0.08775269])

    """
    return vectorize(_bond_yield_0)(
        settle=settle, cpn=cpn, mat=mat, price=price, freq=freq,
        comp_freq=comp_freq, face=face, redeem=redeem, daycount=daycount,
        guess=guess)[()]


def _import_isda_daycounters():
    r"""Import isda_daycounters. Could raise ImportError

    Returns
    -------
    daycounters : dict
         keys are the names of the daycounters, values are the classes
    default_daycounter : daycounter
         thirty360
    no_isda : bool
         Whether isda_daycounters unavailable (False)

    """
    from isda_daycounters import (actual360, actual365,
                                  actualactual, thirty360,
                                  thirtyE360, thirtyE360ISDA)
    daycounters = {x.name: x for x in (actual360, actual365,
                                       actualactual, thirty360,
                                       thirtyE360, thirtyE360ISDA)}
    default_daycounter = thirty360
    no_isda = False
    return daycounters, default_daycounter, no_isda


def _make_simple_day_counter():
    r"""Create a simple daycounter (basically ACT/365)

    Returns
    -------
    daycounters : dict
         only key is 'simple', value is SimpleDayCount
    default_daycounter : daycounter
         SimpleDayCount
    no_isda : bool
         Whether isda_daycounters available (True)

    """
    class SimpleDayCount:
        def day_count(start_date, end_date):
            return (end_date - start_date).days

        def year_fraction(start_date, end_date):
            return (end_date - start_date).days / 365.0

    daycounters = {'simple': SimpleDayCount}
    default_daycounter = SimpleDayCount
    no_isda = True
    from warnings import warn
    warn("Module isda_daycounters is not installed.\n"
         "Only 'simple' daycount (basically ACT/365) is available.\n"
         "To use other daycounts, install isda_daycounters from\n"
         "https://github.com/miradulo/isda_daycounters")
    return daycounters, default_daycounter, no_isda


try:
    daycounters, default_daycounter, no_isda = _import_isda_daycounters()
except ImportError:
    daycounters, default_daycounter, no_isda = _make_simple_day_counter()



"""Utility functions

`newton_wrapper` is a wrapper for scipy.optimize.newton to return root or nan

`edate` moves date(s) by specified no of months (similar to excel edate)

`dataframe_like_dict` creates pandas DataFrame from dict like arguments

"""
import pandas as pd
import numpy as np
from bond_pricing.no_scipy_workarounds import my_irr, no_scipy


def newton_wrapper(f, guess, warn=True):
    r"""Wrapper for `scipy.optimize.newton` to return root or `nan`

    If scipy cannot be imported, falls back on the my_irr function
    that uses root bracketing and bisection.

    Parameters
    ----------
    f : callable
        The function whose zero is to be found
    guess : float
        An initial estimate of the zero somewhere near the actual zero.
    warn : bool, Optional
        If true, a warning is issued when returning nan.
        This happens when `scipy.optimize.newton` does not converge.

    Returns
    -------
    float
         The root if found, numpy.nan otherwise.
    Examples
    --------
    >>> newton_wrapper(lambda x: x**2 - x, 0.8)
    1.0
    >>> newton_wrapper(lambda x: x**2 + 1, 1, warn=False)
    nan
    """
    try:
        from scipy.optimize import newton
    except ImportError:
        no_scipy.warn("newton")
        return my_irr(f, guess=guess, warn=warn)
    root, status = newton(f, guess, full_output=True, disp=False)
    if status.converged:
        return root
    else:
        if warn:
            from warnings import warn
            warn("Newton root finder did not converge. Returning nan")
        return np.nan


def _edate_0(dt, m):
    """Unvectorized form of edate

    Please see its documentation

    Examples
    --------

    >>> _edate_0('2020-01-31', 1)
    Timestamp('2020-02-29 00:00:00')
    """
    return pd.to_datetime(dt) + pd.tseries.offsets.DateOffset(months=m)


_edate = np.vectorize(_edate_0)


def edate(dt, m):
    """Move date(s) by specified no of months (similar to excel edate)

    Parameters
    ----------
    dt : date, object convertible to date or sequence
         The date(s) to be shifted
    m : int or sequence
         Number of months to shift by

    Returns
    -------
    date
        dt shifted by m months

    Examples
    --------

    >>> edate('2020-01-31', 1)
    Timestamp('2020-02-29 00:00:00')


    >>> edate(['2020-01-31', '2020-03-31'], [1, 6])
    array([Timestamp('2020-02-29 00:00:00'), Timestamp('2020-09-30 00:00:00')],
          dtype=object)
    """
    return _edate(dt, np.array(m))[()]


def dict_to_dataframe(d):
    lengths = [1]
    lengths += [len(x) for x in d.values()
                if hasattr(x, "len")
                and not isinstance(x, str)]
    lengths += [x.shape[0] for x in d.values()
                if hasattr(x, "shape")
                and len(x.shape) > 0]
    n = max(lengths)
    return pd.DataFrame(d, index=range(n))



"""Present value module in bond_pricing

This module includes `npv`, `irr` and `duration` for
arbitrary cash flows. The functions for annuities include
`annuity_pv`, `annuity_pv`, `annuity_rate`, `annuity_instalment`,
`annuity_instalment_breakup`, `annuity_periods`. The helper functions
`pvaf` and `fvaf` are intended for use by the annuity functions, but
can be used directly if desired.

"""
import numpy as np
from numpy import array, log, exp, where, vectorize
from bond_pricing.utils import newton_wrapper, dict_to_dataframe


def pvaf(r, n):
    """ Compute Present Value of Annuity Factor

    Parameters
    ----------
    r : float or sequence of floats
        per period interest rate in decimal
    n : float or sequence of floats
        number of periods

    Returns
    -------
    float or array of floats
       The present value of annuity factor

    Examples
    --------
    >>> pvaf(r=0.1, n=10)
    6.144567105704685

    >>> pvaf(10e-2, [5, 10])
    array([3.79078677, 6.14456711])

    """
    r, n = array(r), array(n)
    # We use numpy.where to handle r == 0, but numpy.where evaluates
    # both expressions and so 0/0 will still happen. Suppress the error.
    old_settings = np.seterr(invalid='ignore')
    result = where(r == 0, n, np.divide(1 - (1+r)**-n, r))
    np.seterr(**old_settings)
    return result[()]


def fvaf(r, n):
    """ Compute Future Value of Annuity Factor

    Parameters
    ----------
    r : float or sequence of floats
        per period interest rate in decimal
    n : float or sequence of floats
        number of periods

    Returns
    -------
    float or array of floats
       The future value of annuity factor

    Examples
    --------
    >>> fvaf(r=0.1, n=10)
    15.937424601000023
    >>> fvaf(r=[0, 0.1], n=10)
    array([10.       , 15.9374246])

    """
    r, n = array(r), array(n)
    old_settings = np.seterr(invalid='ignore')
    result = where(r == 0, n, np.divide((1+r)**n - 1, r))
    np.seterr(**old_settings)
    return result[()]


def npv(cf, rate, cf_freq=1, comp_freq=1, cf_t=None,
        immediate_start=False):
    r"""NPV of a sequence of cash flows

    Parameters
    ----------
    cf : float or sequence of floats
         array of cash flows
    rate : float or sequence of floats
         discount rate
    cf_freq : float or sequence of floats, optional
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats, optional
         compounding frequency (for example, 2 for semi-annual)
    cf_t : float or sequence of floats or None, optional
         The timing of cash flows.
         If None, equally spaced cash flows are assumed
    immediate_start : bool or sequence of bool, optional
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.

    Returns
    -------
    float or array of floats
       The net present value of the cash flows

    Examples
    --------
    >>> npv(cf=[-100, 150, -50, 75], rate=5e-2)
    59.327132213429586

    >>> npv(cf=[-100, 150, -50, 75], rate=5e-2, comp_freq=[1, 2])
    array([59.32713221, 59.15230661])

    >>> npv(cf=[-100, 150, -50, 75], rate=5e-2,
    ...     immediate_start=[False, True])
    array([59.32713221, 62.29348882])

    >>> npv(cf=[-100, 150, -50, 75], cf_t=[0, 2, 5, 7], rate=[5e-2, 8e-2])
    array([50.17921321, 38.33344284])

    """

    def one_npv(rate, cf_freq, comp_freq, immediate_start):
        if cf_t is None:
            start = 0 if immediate_start else 1/cf_freq
            stop = start + len(cf) / cf_freq
            cf_ta = np.arange(start=start, step=1/cf_freq, stop=stop)
        else:
            cf_ta = array(cf_t)
        cc_rate = equiv_rate(rate, from_freq=comp_freq, to_freq=np.inf)
        df = exp(-cc_rate * cf_ta)
        return np.dot(cf, df)

    cf = array(cf)
    return vectorize(one_npv)(
        rate=rate, cf_freq=cf_freq, comp_freq=comp_freq,
        immediate_start=immediate_start)[()]


def equiv_rate(rate, from_freq=1, to_freq=1):
    r"""Convert interest rate from one compounding frequency to another

    Parameters
    ----------
    rate : float or sequence of floats
           discount rate in decimal
    from_freq : float or sequence of floats
                compounding frequency of input rate
    to_freq : float or sequence of floats
              compounding frequency of output rate

    Returns
    -------
    float or array of floats
       The discount rate for the desired compounding frequency

    Examples
    --------
    >>> equiv_rate(
    ...    rate=10e-2, from_freq=1, to_freq=[1, 2, 12, 365, np.inf])
    array([0.1       , 0.0976177 , 0.09568969, 0.09532262, 0.09531018])

    >>> equiv_rate(
    ...    rate=10e-2, from_freq=[1, 2, 12, 365, np.inf], to_freq=1)
    array([0.1       , 0.1025    , 0.10471307, 0.10515578, 0.10517092])

    """
    rate, from_freq, to_freq = array(rate), array(from_freq), array(to_freq)
    # the use of where prevents nan from being returned for 0/0
    # but since where evaluates both expressions the error still occurs
    # we run np.seterr and np.divide to catch 0/0 errors
    old_settings = np.seterr(invalid='ignore')
    cc_rate = where(from_freq == np.inf, rate,
                    log(1 + np.divide(rate, from_freq)) * from_freq)
    res = where(from_freq == to_freq,
                rate,
                where(to_freq == np.inf,
                      cc_rate,
                      (exp(np.divide(cc_rate, to_freq)) - 1) * to_freq))[()]
    np.seterr(**old_settings)
    return res


def duration(cf, rate, cf_freq=1, comp_freq=1, cf_t=None,
             immediate_start=False, modified=False):
    r"""Duration of arbitrary sequence of cash flows

    Parameters
    ----------
    cf : sequence of floats
         array of cash flows
    rate : float or sequence of floats
         discount rate
    cf_freq : float or sequence of floats, optional
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats, optional
         compounding frequency (for example, 2 for semi-annual)
    cf_t : float or sequence of floats or None, optional
         The timing of cash flows.
         If None, equally spaced cash flows are assumed
    immediate_start : bool or sequence of bool, optional
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    modified : bool or sequence of bool, optional
         If True, modified duration is returned

    Returns
    -------
    float or array of floats
       The duration of the cash flows

    Examples
    --------
    >>> duration(cf=[100, 50, 75, 25], rate=10e-2)
    1.9980073065426769

    >>> duration(cf=[100, 50, 75, 25], rate=10e-2,
    ...          immediate_start=[True, False])
    array([0.99800731, 1.99800731])

    """

    def one_duration(rate, cf_freq, comp_freq, immediate_start):
        if cf_t is None:
            start = 0 if immediate_start else 1/cf_freq
            stop = start + len(cf) / cf_freq
            cf_ta = np.arange(start=start, step=1/cf_freq, stop=stop)
        else:
            cf_ta = cf_t
        cc_rate = equiv_rate(rate, from_freq=comp_freq, to_freq=np.inf)
        df = exp(-cc_rate * cf_ta)
        return np.dot(cf*df, cf_ta) / np.dot(cf, df)

    D = vectorize(one_duration)(
        rate=rate, cf_freq=cf_freq, comp_freq=comp_freq,
        immediate_start=immediate_start)
    D /= where(modified, 1 + rate/comp_freq, 1)
    return D[()]


def irr(cf, cf_freq=1, comp_freq=1, cf_t=None, r_guess=10e-2):
    r"""IRR of a sequence of cash flows

    Multiple IRRs can be found by giving multiple values of r_guess
    as shown in one of the examples below.

    Parameters
    ----------
    cf : float or sequence of floats
         array of cash flows
    cf_freq : float or sequence of floats, optional
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats, optional
         compounding frequency (for example, 2 for semi-annual)
    cf_t : float or sequence of floats or None, optional
         The timing of cash flows.
         If None, equally spaced cash flows are assumed
    immediate_start : bool or sequence of bool, optional
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    r_guess : float or sequence of floats, optional
         Starting value (guess) for root finder

    Returns
    -------
    float or array of floats
       The internal rate of return (IRR) of the cash flows

    Examples
    --------
    >>> irr(cf=[-100, 150, -50, 75])
    0.4999999999999994

    >>> irr(cf=[-100, 150, -50, 75], cf_freq=1, comp_freq=2)
    0.4494897427831782

    >>> irr(cf=[-100, 150, -50, 75], cf_t=[0, 2, 5, 7])
    0.2247448713915599

    >>> irr(cf=(-100, 230, -132), r_guess=[0.13, 0.18])
    array([0.1, 0.2])

    """
    if np.sign(max(cf)) == np.sign(min(cf)):
        return(np.nan)

    def one_irr(cf_freq, comp_freq, r_guess):

        def f(r):
            return npv(cf=cf, rate=r, cf_freq=cf_freq,
                       comp_freq=comp_freq, cf_t=cf_t)
        return newton_wrapper(f, r_guess)

    return vectorize(one_irr)(
        cf_freq=cf_freq, comp_freq=comp_freq, r_guess=r_guess)[()]


def annuity_pv(rate, n_periods=np.inf, instalment=1, terminal_payment=0,
               immediate_start=False, cf_freq=1, comp_freq=1):
    r"""Present value of annuity

    Parameters
    ----------
    rate : float or sequence of floats
            per period discount rate in decimal
    n_periods : float or sequence of floats
            number of periods of annuity
    instalment : float or sequence of floats
            instalment per period
    terminal_payment : float or sequence of floats
            baloon payment at the end of the annuity
    immediate_start : bool or sequence of bool
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    cf_freq : float or sequence of floats
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats
         compounding frequency (for example, 2 for semi-annual)

    Returns
    -------
    float or array of floats
       The present value of the annuity

    Examples
    --------
    >>> annuity_pv(rate=10e-2, n_periods=15, instalment=500)
    3803.039753154183
    >>> annuity_pv(rate=10e-2, n_periods=[10, 15], instalment=500)
    array([3072.28355285, 3803.03975315])

    """
    rate, n_periods, instalment, terminal_payment, immediate_start = (
        array(rate), array(n_periods), array(instalment),
        array(terminal_payment), array(immediate_start))
    cf_freq, comp_freq = array(cf_freq), array(comp_freq)
    r = equiv_rate(rate, comp_freq, cf_freq) / cf_freq
    pv = pvaf(r, n_periods) * instalment + (
        terminal_payment / (1 + r)**n_periods)
    pv *= where(immediate_start, 1 + r, 1)
    return pv[()]


def annuity_fv(rate, n_periods=np.inf, instalment=1, terminal_payment=0,
               immediate_start=False, cf_freq=1, comp_freq=1):
    r"""Future value of annuity

    Parameters
    ----------
    rate : float or sequence of floats
            per period discount rate in decimal
    n_periods : float or sequence of floats
            number of periods of annuity
    instalment : float or sequence of floats
            instalment per period
    terminal_payment : float or sequence of floats
            baloon payment at the end of the annuity
    immediate_start : bool or sequence of bool
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    cf_freq : float or sequence of floats
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats
         compounding frequency (for example, 2 for semi-annual)

    Returns
    -------
    float or array of floats
       The future value of the annuity

    Examples
    --------
    >>> annuity_fv(rate=10e-2, n_periods=15, instalment=500)
    15886.240847078281
    >>> annuity_fv(rate=10e-2, n_periods=[10, 15], instalment=500)
    array([ 7968.7123005 , 15886.24084708])

    """
    rate, n_periods, instalment, terminal_payment, immediate_start = (
        array(rate), array(n_periods), array(instalment),
        array(terminal_payment), array(immediate_start))
    cf_freq, comp_freq = array(cf_freq), array(comp_freq)
    r = equiv_rate(rate, comp_freq, cf_freq)/cf_freq
    tv = fvaf(r, n_periods) * instalment + terminal_payment
    tv *= where(immediate_start, 1 + r, 1)
    return tv[()]


def annuity_instalment(rate, n_periods=np.inf, pv=None, fv=None,
                       terminal_payment=0, immediate_start=False,
                       cf_freq=1, comp_freq=1):
    r"""Periodic instalment to get desired PV or FV

    Parameters
    ----------
    rate : float or sequence of floats
            per period discount rate in decimal
    n_periods : float or sequence of floats
            number of periods of annuity
    pv : float or sequence of floats
         desired present value of annuity
    fv : float or sequence of floats
         desired future value of annuity
         If pv and fv are given, discounted value of fv is added to pv
    terminal_payment : float or sequence of floats
            baloon payment at the end of the annuity
    immediate_start : bool or sequence of bool
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    cf_freq : float or sequence of floats
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats
         compounding frequency (for example, 2 for semi-annual)

    Returns
    -------
    float or array of floats
       The instalment


    Examples
    --------
    >>> annuity_instalment(rate=10e-2, n_periods=15, pv=3803.04)
    500.0000324537518

    >>> annuity_instalment(rate=10e-2, n_periods=[10, 15], pv=3803.04)
    array([618.92724655, 500.00003245])

    """
    if fv is None:
        pv = pv or 1
        fv = 0
    else:
        pv = pv or 0
    fv, pv, n_periods = array(fv), array(pv), array(n_periods)
    r = equiv_rate(rate, comp_freq, cf_freq)/cf_freq
    reqd_annuity_pv = pv + (fv - terminal_payment) / (1 + r)**n_periods
    return (reqd_annuity_pv / pvaf(r, n_periods))[()]


def annuity_periods(rate, instalment=1, pv=None, fv=None,
                    terminal_payment=0, immediate_start=False,
                    cf_freq=1, comp_freq=1, round2int_digits=6):
    r"""Number of periods of annuity to get desired PV or FV

    Parameters
    ----------
    rate : float or sequence of floats
            per period discount rate in decimal
    instalment : float or sequence of floats
            instalment per period
    pv : float or sequence of floats
         desired present value of annuity
    fv : float or sequence of floats
         desired future value of annuity
         If pv and fv are given, discounted value of fv is added to pv
    terminal_payment : float or sequence of floats
            baloon payment at the end of the annuity
    immediate_start : bool or sequence of bool
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    cf_freq : float or sequence of floats
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats
         compounding frequency (for example, 2 for semi-annual)
    round2int_digits: float or sequence of floats
         answer is rounded to integer if round2int_digits after the
         decimal point are zero

    Returns
    -------
    float or array of floats
       The number of periods

    Examples
    --------
    >>> annuity_periods(rate=10e-2, instalment=500, pv=3803.04)
    15.000002163748604

    >>> annuity_periods(rate=10e-2, instalment=500, pv=3803.04,
    ...      round2int_digits=4)
    15.0

    >>> annuity_periods(rate=[0, 10e-2], instalment=500, pv=3803.04)
    array([ 7.60608   , 15.00000216])

"""
    if fv is None:
        pv = pv or 1
        fv = 0
    else:
        pv = pv or 0
    fv, pv, instalment, terminal_payment = (
        array(fv), array(pv), array(instalment), array(terminal_payment))
    # if immediate_start
    # makes it a deferred annuity (with one less period)
    pv -= where(immediate_start, instalment, 0)
    # pv + fv * df
    #   = terminal_payment * df + instalment * (1 - df) / r
    # pv + fv * df
    #   = terminal_payment * df + instalment / r  - instalment * df / r
    # fv * df - terminal_payment * df + instalment * df / r
    #   =  instalment / r  - pv
    # df * (fv - terminal_payment  + instalment / r)
    #   =  instalment / r  - pv
    r = equiv_rate(rate, comp_freq, cf_freq) / cf_freq
    # We use numpy.where to handle r == 0 or df == 0
    # but numpy.where evaluates both expressions and so
    # we must suppress the error.
    old_settings = np.seterr(divide='ignore', invalid='ignore')
    perpetuity = instalment / r
    df = np.divide(perpetuity - pv, fv - terminal_payment + perpetuity)
    n = where(r == 0,
              (pv + fv) / instalment,
              where((df < 0) | (df > 1),
                    np.nan,
                    where(df == 0,
                          np.inf,
                          -log(df) / log(1 + r))))
    np.seterr(**old_settings)
    # if immediate_start
    # add back the one period that we removed
    n += where(immediate_start, 1, 0)
    # if the result is close to an integer, then round to the integer
    # the tolerance for this is given by round2int_digits
    return where(np.abs(n - n.round(0)) < 10**-round2int_digits,
                 n.round(0).astype(int), n)[()]


def annuity_rate(n_periods=np.inf, instalment=1, pv=None, fv=None,
                 terminal_payment=0, immediate_start=False,
                 cf_freq=1, comp_freq=1, r_guess=0):
    r"""Discount rate to get desired PV or FV of annuity

    Parameters
    ----------
    n_periods : float or sequence of floats
            number of periods of annuity
    instalment : float or sequence of floats
            instalment per period
    pv : float or sequence of floats
         desired present value of annuity
    fv : float or sequence of floats
         desired future value of annuity
         If pv and fv are given, discounted value of fv is added to pv
    terminal_payment : float or sequence of floats
            baloon payment at the end of the annuity
    immediate_start : bool or sequence of bool
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    cf_freq : float or sequence of floats
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats
         compounding frequency (for example, 2 for semi-annual)
    r_guess : float, optional
         Starting value (guess) for root finder

    Returns
    -------
    float or array of floats
       The discount rate

    Examples
    --------
    >>> annuity_rate(n_periods=15, instalment=500, pv=3803.04)
    0.09999998862890495

    >>> annuity_rate(n_periods=[9, 10, 15], instalment=100, pv=1000)
    array([-0.0205697 ,  0.        ,  0.05556497])

    """
    if fv is None:
        pv = pv or 1
        fv = 0
    else:
        pv = pv or 0

    def one_rate(n_periods, instalment, pv, terminal_payment,
                 immediate_start, cf_freq, comp_freq):
        if (n_periods == np.inf):
            return (pv/instalment)

        def f(r):
            return annuity_pv(
                r, n_periods=n_periods, instalment=instalment,
                terminal_payment=terminal_payment,
                immediate_start=immediate_start, cf_freq=cf_freq,
                comp_freq=comp_freq) - pv

        return newton_wrapper(f, r_guess)

    return vectorize(one_rate)(
        n_periods=n_periods, instalment=instalment, pv=pv,
        terminal_payment=terminal_payment,
        immediate_start=immediate_start, cf_freq=cf_freq,
        comp_freq=comp_freq)[()]


def annuity_instalment_breakup(
        rate, n_periods=np.inf, pv=None, fv=None,
        terminal_payment=0, immediate_start=False,
        cf_freq=1, comp_freq=1, period_no=1,
        return_dataframe=False):
    r"""Break up instalment into principal and interest parts

    Parameters
    ----------
    rate : float or sequence of floats
            per period discount rate in decimal
    n_periods : float or sequence of floats
            number of periods of annuity
    pv : float or sequence of floats
         desired present value of annuity
    fv : float or sequence of floats
         desired future value of annuity
         If pv and fv are given, discounted value of fv is added to pv
    terminal_payment : float or sequence of floats
            baloon payment at the end of the annuity
    immediate_start : bool or sequence of bool
         If True, cash flows start immediately
         Else, the first cash flow is at the end of the first period.
    cf_freq : float or sequence of floats
         cash flow frequency (for example, 2 for semi-annual)
    comp_freq : float or sequence of floats
         compounding frequency (for example, 2 for semi-annual)
    return_dataframe : bool
         whether to return pandas DataFrame instead of dict

    Returns
    -------
    dict
            Opening Principal

            Instalment

            Interest Part

            Principal Part

            Closing Principal

    Examples
    --------
    >>> annuity_instalment_breakup(rate=10e-2, n_periods=15, pv=3803.04,
    ...       period_no=6)  # doctest: +NORMALIZE_WHITESPACE
    {'Period No': 6,
     'Opening Principal': 3072.283752266599,
     'Instalment': 500.0000324537518,
     'Interest Part': 307.2283752266599,
     'Principal Part': 192.7716572270919,
     'Closing Principal': 2879.512095039507}

    >>> d = annuity_instalment_breakup(rate=10e-2, n_periods=15, pv=3803.04,
    ...       period_no=range(1, 4), return_dataframe=True
    ...       ); print(d.iloc[:, :4]); print(d.iloc[:, 4:])
       Period No  Opening Principal  Instalment  Interest Part
    0          1        3803.040000  500.000032     380.304000
    1          2        3683.343968  500.000032     368.334397
    2          3        3551.678332  500.000032     355.167833
       Principal Part  Closing Principal
    0      119.696032        3683.343968
    1      131.665636        3551.678332
    2      144.832199        3406.846133

    """
    fv, pv, n_periods = array(fv), array(pv), array(n_periods)
    cf_freq, comp_freq = array(cf_freq), array(comp_freq)
    period_no = array(period_no)
    assert (np.all(period_no > 0) and
            np.all(period_no > 0) and
            np.all(period_no <= n_periods) and
            np.all(period_no.astype(int) == period_no)
            ), "Invalid period_no"
    instalment = annuity_instalment(
        rate=rate, n_periods=n_periods, pv=pv,
        immediate_start=immediate_start, cf_freq=cf_freq,
        comp_freq=comp_freq)
    r = equiv_rate(rate, comp_freq, cf_freq)/cf_freq
    df = (1 + r)**(period_no-1)
    opening_principal = pv * df - annuity_fv(
        rate=rate, n_periods=period_no-1, instalment=instalment,
        immediate_start=immediate_start,
        cf_freq=cf_freq, comp_freq=comp_freq)
    result = {"Period No": period_no[()],
              "Opening Principal": opening_principal,
              "Instalment": instalment,
              "Interest Part": opening_principal * r,
              "Principal Part": instalment - opening_principal * r,
              "Closing Principal":
              opening_principal + opening_principal * r - instalment}
    if return_dataframe:
        return dict_to_dataframe(result)
    else:
        return result



from numpy import (array, arange, empty_like, vectorize,  # noqa E401
                   ceil, log, exp, interp, nan, where, dot,
                   concatenate)
import numpy as np
from bond_pricing.simple_bonds import bond_coupon_periods, equiv_rate
from bond_pricing.utils import newton_wrapper, dict_to_dataframe
from bond_pricing.no_scipy_workarounds import no_scipy


def par_yld_to_zero(par, freq=1, return_dataframe=False):
    r"""Bootstrap a complete par bond yield curve to zero

    Parameters
    ----------
    par : sequence of floats
          The par bond yields for various maturities in decimal
          Maturities are spaced 1/freq years apart
    freq : int, optional
          The coupon frequency (equals compounding frequency)
    return_dataframe : bool, optional
         whether to return pandas DataFrame instead of dict
    Returns
    -------
    dict or DataFrame:

          zero_yields: array of zero yields in decimal

          zero_prices: array of zero prices

          forward_rates: array of forward rates in decimal

          (Maturities are spaced 1/freq years apart)

    Examples
    --------
    >>> par_yld_to_zero(
    ... par=[1.0200e-2, 1.2000e-2, 1.4200e-2, 1.6400e-2, 1.9150e-2,
    ...      2.1900e-2, 2.4375e-2, 2.6850e-2, 2.9325e-2, 3.1800e-2],
    ... freq=2, return_dataframe=True)
       zero_yields  zero_prices  forward_rates
    0     0.010200     0.994926       0.010200
    1     0.012005     0.988102       0.013812
    2     0.014220     0.978970       0.018656
    3     0.016445     0.967776       0.023133
    4     0.019245     0.953245       0.030487
    5     0.022068     0.936279       0.036242
    6     0.024630     0.917891       0.040066
    7     0.027218     0.897504       0.045429
    8     0.029837     0.875223       0.050915
    9     0.032492     0.851159       0.056545

    """
    annuity = 0
    prev_zero = 1
    zp = empty_like(par)
    zyld = empty_like(par)
    fwd = empty_like(par)
    for i, cpn in enumerate(par):
        n = i + 1
        pv_intermediate = annuity * cpn/freq
        pv_final = 1 - pv_intermediate
        zero_price = pv_final / (1 + cpn/freq)
        zp[i] = zero_price
        zyld[i] = (zero_price**(-1.0/n) - 1) * freq
        fwd[i] = (prev_zero / zero_price - 1) * freq
        prev_zero = zero_price
        annuity += zero_price
    result = dict(zero_yields=zyld, zero_prices=zp, forward_rates=fwd)
    if return_dataframe:
        return dict_to_dataframe(result)
    else:
        return result


def zero_to_par(zero_prices=None, zero_yields=None, freq=1):
    r"""Convert zero yield curve into par curve

    The zero curve can be specified either as prices or as yields.
    One of `zero_prices` and `zero_yields` must not be None.

    Parameters
    ----------
    zero_prices : sequence of floats, optional
             The zero prices for various maturities in decimal
             Either zero_prices or zero_yields must be provided
             If both are provided, zero_yields is ignored
             Maturities are spaced 1/freq years apart
    zero_yields : sequence of floats, optional
             The zero yields for various maturities in decimal
             Maturities are spaced 1/freq years apart
    freq : int, optional
          The coupon frequency (equals compounding frequency)

    Returns
    -------
    par : array of floats
          The par yields for various maturities in decimal
          Maturities are spaced 1/freq years apart

    Examples
    --------
    >>> zero_to_par(
    ... zero_yields=[0.0102    , 0.0120054 , 0.01421996, 0.01644462,
    ...              0.01924529, 0.02206823, 0.02462961, 0.02721789,
    ...              0.0298373 , 0.0324924 ],
    ... freq=2)  # doctest: +NORMALIZE_WHITESPACE
    array([0.0102  , 0.012   , 0.0142  , 0.0164  , 0.01915 , 0.0219  ,
           0.024375, 0.02685 , 0.029325, 0.0318  ])

    >>> zero_to_par(
    ... zero_prices=[0.99492588, 0.98810183, 0.97896982, 0.96777586,
    ...              0.9532451 , 0.9362787 , 0.91789052, 0.89750426,
    ...              0.87522337, 0.85115892],
    ... freq=2)  # doctest: +NORMALIZE_WHITESPACE
    array([0.0102  , 0.012   , 0.0142  , 0.0164  , 0.01915 , 0.0219  ,
           0.024375, 0.02685 , 0.029325, 0.0318  ])

    """
    if zero_prices is None:
        zero_yields, freq = array(zero_yields), array(freq)
        t = arange(len(zero_yields)) + 1
        zero_prices = (1 + zero_yields/freq)**(-t)
    annuity = 0
    par = empty_like(zero_prices)
    for i, zero_price in enumerate(zero_prices):
        annuity = annuity + zero_price
        pv_redeem = zero_price
        pv_cpns = 1 - pv_redeem
        par[i] = (pv_cpns / annuity) * freq
    return par


def nelson_siegel_zero_rate(beta0, beta1, beta2, tau, m):
    r"""Computes zero rate from Nelson Siegel parameters

    Parameters
    ----------
    beta0 : float
            the long term rate
    beta1 : float
            the slope
    beta2 : float
            curvature
    tau : float
            location of the hump
    m : float
            maturity at which zero rate is to be found

    Returns
    -------
    float :
        the continuously compounded zero yield for maturity m

    Examples
    --------
    >>> nelson_siegel_zero_rate(0.128397, -0.024715, -0.050231, 2.0202,
    ...                         [0.25, 5, 15, 30])
    array([0.10228692, 0.10489195, 0.11833924, 0.12335016])

    >>> nelson_siegel_zero_rate(0.0893088, -0.0314768, -0.0130352,
    ...                         3.51166, [0.25, 5, 15, 30])
    array([0.05848376, 0.06871299, 0.07921554, 0.08410199])

    """
    m = array(m)
    old_settings = np.seterr(invalid='ignore')
    possible_0_by_0 = where(m == 0,
                            1,
                            np.divide(1 - exp(-m/tau), m/tau))
    np.seterr(**old_settings)
    return (beta0 + (beta1 + beta2) * possible_0_by_0
            - beta2 * exp(-m/tau))[()]


def make_zero_price_fun(flat_curve=None,
                        nelson_siegel=None,
                        par_at_knots=None,
                        par_at_coupon_dates=None,
                        zero_at_knots=None,
                        zero_at_coupon_dates=None,
                        freq=1, max_maturity=None):
    r"""Create function that returns zero price for any maturity

    The zero curve can be specified in many alternative ways:
    `flat_curve`, `nelson_siegel`, `par_at_knots`,
    `par_at_coupon_dates`, `zero_at_knots`, `zero_at_coupon_dates`.
    One of these must not be None.


    Parameters
    ----------
    flat_curve : float, optional
           Yield (flat yield curve)
    nelson_siegel : tuple of floats, optional
           tuple consists of beta0, beta1, beta2, tau
    par_at_knots : tuple of two sequences of floats, optional
           First element of tuple is a sequence of maturities
           Second element is a sequence of par yields for these maturities
    par_at_coupon_dates : sequence of floats, optional
           Par yields for maturities spaced 1/freq years apart
    zero_at_knots : tuple of two sequences of floats, optional
           First element of tuple is a sequence of maturities
           Second element is a sequence of zero rates for these maturities
    zero_at_coupon_dates : sequence, optional
          Zero yields for maturities spaced 1/freq years apart
    freq : int, optional
          The coupon frequency (equals compounding frequency)
    max_maturity : float, optional
          The maximum maturity upto which the yields are to be
          extrapolated. If None, no extrapolation is done.

    Returns
    -------
    function :
          Function that takes float (maturity) as argument and
          returns float (zero price)

    Examples
    --------
    >>> make_zero_price_fun(par_at_knots = (
    ... [0, 1, 3, 5, 10],
    ... [3e-2, 3.5e-2,  4e-2, 4.5e-2, 4.75e-2]))([1, 5, 10])
    array([0.96618357, 0.80065643, 0.62785397])

    """
    if flat_curve is not None:
        r = equiv_rate(flat_curve, freq, np.inf)
        return lambda t: exp(- r * array(t))[()]
    if nelson_siegel is not None:
        beta0, beta1, beta2, tau = nelson_siegel
        return lambda m: exp(- nelson_siegel_zero_rate(
            beta0, beta1, beta2, tau, m) * array(m))[()]
    if par_at_knots is not None:
        t, r = par_at_knots
        if min(t) != 0:
            from warnings import warn
            if min(t) < 0:
                warn("Knot point at negative maturity.")
            else:
                warn("A knot point at zero maturity is recommended.")
        if max_maturity is None:
            max_maturity = max(t)
        par_at_coupon_dates = interpolate(
            t, r, arange(ceil(max_maturity*freq))+1)
    if par_at_coupon_dates is not None:
        zero_at_coupon_dates = par_yld_to_zero(par_at_coupon_dates,
                                               freq)['zero_yields']
    if zero_at_coupon_dates is None:
        assert zero_at_knots is not None, "No yield curve provided"
        t, r = zero_at_knots
        if min(t) != 0:
            from warnings import warn
            if min(t) < 0:
                warn("Knot point at negative maturity.")
            else:
                warn("A knot point at zero maturity is recommended.")
        if max_maturity is None:
            max_maturity = max(t)
        zero_at_coupon_dates = interpolate(
            t, r, arange(ceil(max_maturity*freq))+1)
    zero_at_coupon_dates = equiv_rate(zero_at_coupon_dates,
                                      from_freq=freq, to_freq=np.inf)
    t = arange(len(zero_at_coupon_dates) + 1) / freq
    log_zero_df = -concatenate([[0], zero_at_coupon_dates]) * t
    return lambda x: exp(interp(x, t, log_zero_df, left=nan, right=nan))


def zero_curve_bond_price_breakup(
        settle=None, cpn=0, mat=1, zero_price_fn=(lambda x: 1),
        freq=2, face=100, redeem=None, daycount=None,
        return_dataframe=False):
    r"""Clean/dirty price & accrued interest of coupon bond using zero yields

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    zero_price_fn : function
          takes float (maturity) as argument and
          returns float (zero price)
    freq : int
          Coupon frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods
    return_dataframe : bool
         whether to return pandas DataFrame instead of dict

    Returns
    -------
    dict or DataFrame
         dirty: dirty price of the bond

         accrued: accrued interest

         clean: clean price of the bond

         next_coupon: Next coupon date (only if settle is None)

         prev_coupon: Previous coupon date (only if settle is None)

    Examples
    --------
    >>> zero_curve_bond_price_breakup(
    ...     cpn=10e-2, mat=10, freq=1,
    ...     zero_price_fn=make_zero_price_fun(
    ...         flat_curve=8e-2))  # doctest: +NORMALIZE_WHITESPACE
    {'DirtyPrice': 113.42016279788285,
     'AccruedInterest': 0.0,
     'CleanPrice': 113.42016279788285,
     'NextCoupon': None,
     'PreviousCoupon': None}

    >>> zero_curve_bond_price_breakup(
    ...     cpn=5e-2, mat=2, freq=1,
    ...     zero_price_fn=make_zero_price_fun(
    ...         zero_at_coupon_dates=[3e-2, 10e-2])
    ... )  # doctest: +NORMALIZE_WHITESPACE
    {'DirtyPrice': 91.63122843617106,
     'AccruedInterest': 0.0,
     'CleanPrice': 91.63122843617106,
     'NextCoupon': None,
     'PreviousCoupon': None}

    """
    freq, cpn = array(freq), array(cpn)
    redeem = where(redeem is None, face, redeem)
    red_by_face = redeem / face
    res = bond_coupon_periods(settle, mat, freq, daycount)
    # print(res)

    def one_dirty_price(n, fraction, coupon, R_by_F, zp_fun):
        t = (arange(n) + fraction) / freq
        df = zp_fun(t)
        cf = [coupon/freq] * int(n-1) + [coupon/freq + R_by_F]
        # import pandas as pd
        # print(pd.DataFrame(dict(df=df, cf=cf), index=t))
        return dot(df, cf)

    dirty_price = vectorize(one_dirty_price)
    dirty = dirty_price(n=res['n'],
                        fraction=res['discounting_fraction'],
                        coupon=cpn,
                        R_by_F=red_by_face,
                        zp_fun=zero_price_fn)
    accrued = cpn/freq * res['accrual_fraction']
    clean = dirty - accrued
    result = dict(DirtyPrice=(face*dirty)[()],
                  AccruedInterest=(face*accrued)[()],
                  CleanPrice=(face*clean)[()],
                  NextCoupon=res['next_coupon'],
                  PreviousCoupon=res['prev_coupon'])
    if return_dataframe:
        return dict_to_dataframe(result)
    else:
        return result


def zero_curve_bond_price(settle=None, cpn=0, mat=1,
                          zero_price_fn=(lambda x: 1),
                          freq=2, face=100, redeem=None,
                          daycount=None):
    r"""Compute clean price of coupon bond using zero yields

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    zero_price_fn : function
          takes float (maturity) as argument and
          returns float (zero price)
    freq : int
          Coupon frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods

    Returns
    -------
    float
         clean price of the bond

    Examples
    --------
    >>> zero_curve_bond_price(
    ...     cpn=10e-2, mat=10, freq=1,
    ...     zero_price_fn=make_zero_price_fun(
    ...         flat_curve=8e-2))  # doctest: +NORMALIZE_WHITESPACE
    113.42016279788285

    >>> zero_curve_bond_price(
    ...     cpn=5e-2, mat=2, freq=1,
    ...     zero_price_fn=make_zero_price_fun(
    ...         zero_at_coupon_dates=[3e-2, 10e-2])
    ... )  # doctest: +NORMALIZE_WHITESPACE
    91.63122843617106

    >>> zero_curve_bond_price(
    ...     cpn=5.792982e-2, mat=6, freq=2,
    ...     zero_price_fn=make_zero_price_fun(
    ...         nelson_siegel=(6.784e-2, -3.8264e-2, -3.6631e-2, 0.7774))
    ... )
    99.99999939355965

    """
    return zero_curve_bond_price_breakup(
        settle=settle, cpn=cpn, mat=mat, zero_price_fn=zero_price_fn,
        freq=freq, face=face, redeem=redeem,
        daycount=daycount)['CleanPrice']


def zero_curve_bond_duration(settle=None, cpn=0, mat=1,
                             zero_price_fn=(lambda x: 1),
                             freq=2, face=100, redeem=None,
                             daycount=None, modified=False):
    r"""Duration of coupon bond using zero yields

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    zero_price_fn : function
          takes float (maturity) as argument and
          returns float (zero price)
    freq : int
          Coupon frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods

    Returns
    -------
    float
         duration or modified duration of the bond

    Examples
    --------
    >>> zero_curve_bond_duration(
    ...     cpn=10e-2, mat=10, freq=1,
    ...     zero_price_fn=make_zero_price_fun(
    ...         flat_curve=8e-2))  # doctest: +NORMALIZE_WHITESPACE
    6.965803939497351
    >>> zero_curve_bond_duration(
    ...     cpn=5e-2, mat=2, freq=1,
    ...     zero_price_fn=make_zero_price_fun(
    ...         zero_at_coupon_dates=[3e-2, 10e-2])
    ... )  # doctest: +NORMALIZE_WHITESPACE
    1.9470227670753064
    """
    freq, cpn = array(freq), array(cpn)
    redeem = where(redeem is None, face, redeem)
    red_by_face = redeem / face
    res = bond_coupon_periods(settle, mat, freq, daycount)

    def one_duration(n, fraction, coupon, R_by_F, zp_fun):
        t = (arange(n) + fraction) / freq
        df = zp_fun(t)
        cf = [coupon/freq] * int(n-1) + [coupon/freq + R_by_F]
        # import pandas as pd
        # print(pd.DataFrame(dict(df=df, cf=cf), index=t))
        return dot(df, cf * t) / dot(df, cf)

    duration = vectorize(one_duration)
    result = duration(n=res['n'],
                      fraction=res['discounting_fraction'],
                      coupon=cpn,
                      R_by_F=red_by_face,
                      zp_fun=zero_price_fn)
    return result[()]


def static_zero_spread(settle=None, cpn=0, mat=1, price=None,
                       zero_price_fn=(lambda x: 1),
                       freq=2, face=100, redeem=None,
                       daycount=None, guess=0.0):
    r"""Static spread (Z-spread) over zero curve to match bond price

    Parameters
    ----------
    settle : date or None
          The settlement date. None means maturity is in years.
    cpn : float
          The coupon rate in decimal
    mat : float or date
          Maturity date or if settle is None, maturity in years
    price : float
          Market price of the bond
    zero_price_fn : function
          takes float (maturity) as argument and
          returns float (zero price)
    freq : int
          Coupon frequency
    face : float
          Face value of the bond
    redeem : float
          Redemption value
    daycount : daycounter
          This class has day_count and year_fraction methods
    guess : float
          Initial guess of the yield for the root finder

    Returns
    -------
    float
         Static spread (Z-spread)

    Examples
    --------
    >>> round(static_zero_spread(
    ...     cpn=5e-2, mat=2, freq=1, price=91.63122843617106,
    ...     zero_price_fn=make_zero_price_fun(
    ...         zero_at_coupon_dates=[2.5e-2, 9.5e-2])),
    ... 6)
    0.005
    """

    def zero_price_fn_with_spread(spread, freq):
        def f(t):
            cc_yld = -log(zero_price_fn(t)) / t
            new_yld = equiv_rate(cc_yld, np.inf, freq) + spread
            new_cc_yld = equiv_rate(new_yld, freq, np.inf)
            return exp(-new_cc_yld * t)
        return f

    def one_spread(settle, cpn, mat, price, zero_price_fn,
                   freq, face, redeem, daycount, guess):
        return newton_wrapper(
            lambda spread: zero_curve_bond_price(
                settle=settle, cpn=cpn, mat=mat,
                zero_price_fn=zero_price_fn_with_spread(spread, freq),
                freq=freq, face=face,
                redeem=redeem, daycount=daycount) - price,
            guess)
    spread = vectorize(one_spread)
    result = spread(settle=settle, cpn=cpn, mat=mat, price=price,
                    zero_price_fn=zero_price_fn, freq=freq, face=face,
                    redeem=redeem, daycount=daycount, guess=guess)
    return result[()]


def interpolate(t, r, x):
    try:
        from scipy.interpolate import CubicSpline
    except ImportError:
        no_scipy.warn("CubicSpline")
        return np.interp(x, t, r)
    return CubicSpline(t, r)(x)



from bond_pricing.simple_bonds import (  # noqa E401
    bond_price_breakup, bond_price, bond_duration,
    bond_yield)
from bond_pricing.present_value import (  # noqa E401
    pvaf, fvaf, npv, irr, equiv_rate, duration, annuity_fv, annuity_pv,
    annuity_instalment, annuity_instalment_breakup, annuity_periods,
    annuity_rate)
from bond_pricing.utils import (newton_wrapper, edate)  # noqa E401
from bond_pricing.zero_curve_bond_price import (  # noqa E401
    par_yld_to_zero, zero_to_par, nelson_siegel_zero_rate,
    make_zero_price_fun, zero_curve_bond_price_breakup,
    zero_curve_bond_price, zero_curve_bond_duration,
    static_zero_spread)
from bond_pricing.key_rates import (  # noqa E401
    key_rate_shift, make_KRS,
    standard_krs_points, key_rate_shifted_zero_curve)
from bond_pricing.no_scipy_workarounds import (  # noqa E401
    my_irr, no_scipy)



import numpy as np
import pandas as pd
from bond_pricing.zero_curve_bond_price import (  # noqa F401
    zero_to_par, par_yld_to_zero, make_zero_price_fun,
    zero_curve_bond_price)


standard_krs_points = [2, 5, 10, 30]


def _mat_sequence(T, freq):
    return np.arange(1/freq, T + 1/freq, 1/freq)


def key_rate_shift(krs, mat=None, T=None, freq=2,
                   krs_points=standard_krs_points):
    r"""Shift in various par bond yields for a key rate shift

    Mainly for internal use or pedagogical use.


    Parameters
    ----------
    krs : int or string, optional
          The key rate segment of the curve to be shifted
          For example, 5 for shifting the 5 year segment.
          "all" and "none" are also accepted.
    mat : array
          The maturities for which the shift is to be computed
    T : Maximum maturity upto which shifted yields are to be returned.
    freq : Coupon or Compounding Frequency
    krs_points : sequence, optional
           The key rates to be used.

    Returns
    -------
    array
      The shifts in each par bond yields for a one basis point
      shift in a key rate

    Examples
    --------

    """
    if mat is None:
        if T is None:
            T = krs_points[-1]
        mat = _mat_sequence(T, freq)
    if krs in krs_points:
        bump = [0] * len(krs_points)
        bump[krs_points.index(krs)] = 1
        return np.interp(mat, krs_points, bump)
    elif krs == "all":
        return np.ones_like(mat)
    elif krs == "none":
        return np.zeros_like(mat)
    else:
        raise(Exception("Invalid key rate shift"))


def make_KRS(freq=2, krs_points=standard_krs_points):
    r"""Data Frame of par yield shifts for all key rate shifts

    Mainly for pedagogical use.

    Parameters
    ----------
    freq : Coupon or Compounding Frequency
    krs_points : sequence, optional
           The key rates to be used.

    Returns
    -------
    pandas DataFrame

    Examples
    --------

    """
    KRS = pd.DataFrame(index=_mat_sequence(krs_points[-1], freq))
    for krs in list(krs_points) + ['all', 'none']:
        KRS[krs] = key_rate_shift(krs, freq=freq, krs_points=krs_points)
    return KRS


def key_rate_shifted_zero_curve(
        initial_zero_fn=None, initial_zero_price=None,
        initial_zero_yld=None, initial_par_yld=None,
        krs='none', freq=2, T=None, what='zero_yields',
        krs_points=standard_krs_points):
    r"""Applies a key rate shift to an yield curve

    The initial yield curve can be given in many alternative ways.
    One of these must be non None:
    `initial_zero_fn`, `initial_zero_price`, `initial_zero_yld`,
    `initial_par_yld`

    Parameters
    ----------
    initial_zero_fn : function, optional
                      Function that returns zero prices for any maturity
    initial_zero_price : array, optional
                      Zero prices for coupon dates up to some maturity
    initial_zero_yld : array, optional
                      Zero yields for coupon dates up to some maturity
    initial_par_yld : array, optional
                      Par yields for coupon dates up to some maturity
    krs : int or string, optional
          The key rate segment of the curve to be shifted
          For example, 5 for shifting the 5 year segment.
          "all" and "none" are also accepted.
    freq : Coupon or Compounding Frequency
    T : Maximum maturity upto which shifted yields are to be returned.
    what : str, optional
           What to return. Can be:
           `zero_yields`, `zero_prices`, `forward_rates` or `zero_fn`
    krs_points : sequence, optional
           The key rates to be used.
    Returns
    -------
    array or function:
       If `what` is `zero_yields`, `zero_prices` or `forward_rates`,
       the corresponding array is return. If `what` is `zero_fn`, a
       function is returned.

    Examples
    --------
    Compute KR01 of a 10-year 8% bond when initial yield curve is
    flat at 5%. Being a 10 year bond, the 10 year KR01 is the largest.
    But since it is not a par bond, the 5 year KR01 is also large.

    >>> fn0 = make_zero_price_fun(flat_curve=5e-2, freq=2)
    >>> P0 = zero_curve_bond_price(cpn=8e-2, mat=10, zero_price_fn=fn0)
    >>> fns = [key_rate_shifted_zero_curve(
    ...        initial_zero_fn=fn0, krs=krs, what='zero_fn')
    ...        for krs in standard_krs_points]
    >>> zero_curve_bond_price(cpn=8e-2, mat=10, zero_price_fn=fns) - P0
    array([-0.00121608, -0.00467591, -0.08307928,  0.        ])

    """
    if initial_zero_fn is not None:
        if T is None:
            T = krs_points[-1]
        initial_zero_price = initial_zero_fn(_mat_sequence(T, freq))
    if initial_zero_price is not None:
        initial_par_yld = zero_to_par(zero_prices=initial_zero_price,
                                      freq=freq)
    elif initial_zero_yld is not None:
        initial_par_yld = zero_to_par(zero_yields=initial_zero_yld,
                                      freq=freq)
    assert initial_par_yld is not None, (
        "zero_fn, zero_yld, zero_price or par_yld must be given")
    if not isinstance(initial_par_yld, (np.ndarray, pd.Series)):
        initial_par_yld = np.asarray(initial_par_yld)
    shifted_par = initial_par_yld + key_rate_shift(
        krs, T=len(initial_par_yld)/freq, freq=freq,
        krs_points=krs_points) / 10000
    if what in "zero_yields zero_prices forward_rates".split():
        return par_yld_to_zero(shifted_par, freq=freq)[what]
    elif what == "zero_fn":
        zyld = par_yld_to_zero(shifted_par, freq=freq)['zero_yields']
        return make_zero_price_fun(zero_at_coupon_dates=zyld, freq=freq)
    else:
        raise(Exception("'what' must be one of zero_yields zero_prices "
                        "forward_rates zero_fn"))



from numpy import abs, sign, nan, vectorize


class no_scipy:
    warned = []
    msgs = dict(CubicSpline="Using linear interpolation instead",
                newton="Using root bracketing and bisection")

    @staticmethod
    def warn(fn_name):
        if fn_name not in no_scipy.warned:
            no_scipy.warned += [fn_name]
            from warnings import warn
            if fn_name in no_scipy.msgs:
                msg = no_scipy.msgs[fn_name]
            else:
                msg = ""
            no_scipy_warning = (
                "Could not import {:} from Scipy. {:}".
                format(fn_name, msg))
            warn(no_scipy_warning)


LOWER = -1 + 1e-6  # -100% (theoretical lower bound)
UPPER = 1e4  # 1,000,000% (practical upper bound)


def my_irr_0(f, lower=None, upper=None, guess=None, warn=True,
             toler=1e-6):
    if not guess:
        if lower and upper:
            guess = (lower + upper) / 2
        elif lower:
            guess = lower
        elif upper:
            guess = upper
        else:
            guess = 0
    lower = lower or LOWER
    upper = upper or UPPER
    if not(lower >= -1 and guess >= lower and guess <= upper):
        if warn:
            from warnings import warn
            warn("-1 <= lower <= guess <= upper is required")
        return None
    # we first bracket the root (f has opposite signs at L and R)
    L, R = bracket_root(f, lower, upper, guess)
    if L is None:
        print("failed to bracket root")
        return nan
    # Use bisection to find the root between L and R
    return bisection(f, L, R, toler)


my_irr = vectorize(my_irr_0)


def bracket_root(f, lower, upper, guess, nstep=100):
    def grid_search(step_pct, nstep=100):
        step = step_pct / 100
        guess_sign = sign(f(guess))
        L = None
        R = None
        for n in range(nstep):
            # search for change of sign
            # moving in steps of size step_pct
            # stop if sign change found
            if(guess + n*step <= UPPER and
               (sign(f(guess + n*step)) != guess_sign)):
                R = guess + n*step
                L = R - step
                break
            if(guess - n*step >= LOWER and
               (sign(f(guess - n*step)) != guess_sign)):
                L = guess - n*step
                R = L + step
                break
        return L, R
    for step in [1, 5, 10, 25, 100]:
        # try steps of 1%, 5%, ...
        L, R = grid_search(step)
        if L is not None:
            break
    return L, R


def bisection(f, a, b, tol):
    assert sign(f(a)) != sign(f(b))
    m = (a + b)/2

    if abs(f(m)) < tol:
        # stopping condition, report m as root
        return m
    elif sign(f(a)) == sign(f(m)):
        # case where m is an improvement on a.
        # Make recursive call with a = m
        return bisection(f, m, b, tol)
    elif sign(f(b)) == sign(f(m)):
        # case where m is an improvement on b.
        # Make recursive call with b = m
        return bisection(f, a, m, tol)



#!/usr/bin/env python
from setuptools import setup
from os import environ

dependencies = ['numpy', 'pandas']
if 'no_scipy' not in environ:
    dependencies += ['scipy']

setup(install_requires=dependencies)
